"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var howler_1 = require("howler");
var SoundLoader = (function () {
    function SoundLoader() {
        /* ATTRIBUTES */
        /**
         * The percetange of progression
         * @readonly
         */
        this.progress = 0;
        /**
         * Event to fired during the loading
         * @readonly
         */
        this.onProgressEvent = [];
        /**
         * Sound Manager provided by the loader
         * @readonly
         */
        this.manager = new SoundManager();
        /**
         * List of all sounds to load
         * @readonly
         */
        this.sounds = {};
        /**
         * OnLoad event
         */
        this.onLoad = null;
    }
    /* METHODS */
    /**
     * Add a new sound to load
     * @param id - id of the sound
     * @param url - path of the sound
     */
    SoundLoader.prototype.preload = function (id, url) {
        var sound = new howler_1.Howl({
            src: [].concat(url),
            preload: false
        });
        sound.on("load", this._onSoundLoad.bind(this));
        this.sounds[id] = sound;
    };
    /**
     * Know if the sound has completly load all sounds
     */
    SoundLoader.prototype.isReady = function () {
        var _this = this;
        return !Object.keys(this.sounds).
            map(function (key) { return _this.sounds[key].state(); }).
            filter(function (state) { return state !== "loaded"; }).length;
    };
    /**
     * Load all sounds
     */
    SoundLoader.prototype.load = function (onLoad) {
        var sounds = this.getAll(), onProgress = this._onProgress.bind(this);
        this.onLoad = onLoad;
        sounds.forEach(function (sound) { return sound.once("load", onProgress); });
        sounds.forEach(function (sound) { return sound.load(); });
        this._onProgress();
    };
    /**
     * Get all sounds in array
     */
    SoundLoader.prototype.getAll = function () {
        var _this = this;
        return Object.keys(this.sounds).map(function (key) { return _this.sounds[key]; });
    };
    /**
     * Add an event with a callback fired during the loading
     * @param callback - Function to load during the loading
     */
    SoundLoader.prototype.onProgress = function (callback) {
        this.onProgressEvent.push(callback);
    };
    /* PRIVATE */
    /**
     * Event fired when a sound is loaded
     * @private
     */
    SoundLoader.prototype._onSoundLoad = function () {
        if (this.onLoad && this.isReady()) {
            this.manager.sounds = this.sounds;
            this.onLoad();
        }
    };
    /**
     * Fired when the loading is in progress
     */
    SoundLoader.prototype._onProgress = function () {
        var _this = this;
        var soundsKeys = Object.keys(this.sounds);
        this.progress = soundsKeys.length ? (soundsKeys.filter(function (key) { return _this.sounds[key].seek() >= 0; }).length / soundsKeys.length) * 100 : 100;
        this.onProgressEvent.forEach(function (func) { return func(_this.progress); });
    };
    return SoundLoader;
}());
exports.SoundLoader = SoundLoader;
/**
 * Sound manager (use it directly with Assets.sounds)
 */
var SoundManager = (function () {
    function SoundManager() {
        /* ATTRIBUTES */
        /**
         * current music
         * @readonly
         */
        this.music = null;
        /**
         * List of all sounds availabe
         * @readonly
         */
        this.sounds = {};
        /**
         * Know if the sound is muted
         * @readonly
         */
        this.soundMuted = false;
        /**
         * Know if the music is muted
         * @readonly
         */
        this.musicMuted = false;
    }
    /* METHODS */
    /**
     * Play a sound
     * @param id - id of the sound to play
     */
    SoundManager.prototype.play = function (id) {
        if (this.soundMuted) {
            return null;
        }
        var sound = this.sounds[id];
        if (sound) {
            sound.play();
        }
        return sound;
    };
    /**
     * Play a music background
     * @param id - id of the sound to play
     */
    SoundManager.prototype.playMusic = function (id, loop) {
        if (loop === void 0) { loop = false; }
        if (this.musicMuted) {
            return null;
        }
        if (this.music) {
            this.music.stop();
        }
        if (id) {
            this.music = this.play(id);
        }
        else if (this.music) {
            this.music.play();
        }
        if (loop) {
            this.music.loop(loop);
        }
        return this.music;
    };
    /**
     * Stop the current music
     * @param stopAndRemove - If true, the current will be removed from the buffer
     */
    SoundManager.prototype.stop = function (stopAndRemove) {
        if (stopAndRemove === void 0) { stopAndRemove = false; }
        if (this.music) {
            this.music.stop();
            if (stopAndRemove) {
                this.music = null;
            }
        }
    };
    /**
     * Mute all sounds
     */
    SoundManager.prototype.mute = function (music, sound) {
        if (typeof music === "undefined" && typeof sound === "undefined") {
            howler_1.Howler.mute(true);
            return null;
        }
        if (music) {
            this.musicMuted = true;
            this.stop();
        }
        if (sound) {
            this.soundMuted = true;
        }
    };
    /**
     * Unmute sounds
     */
    SoundManager.prototype.unmute = function (music, sound) {
        if (typeof music === "undefined" && typeof sound === "undefined") {
            howler_1.Howler.mute(false);
            return null;
        }
        if (music) {
            this.musicMuted = false;
            this.playMusic();
        }
        if (sound) {
            this.soundMuted = false;
        }
    };
    return SoundManager;
}());
exports.SoundManager = SoundManager;
